﻿/**
* @file Logging.h
* @author Iacopo Masi
* @date 2010-05-10
*/

#ifndef LOGGING_H_
#define LOGGING_H_

#define LOGGING


#include <fstream>
#include <opencv/cv.h>
#include <map>
#include <iostream>
#include <string>

using namespace std;



class Log{
public:
	ofstream file;

	static Log* getInstance(string keyname){
	
		if (instances[keyname] == NULL)
			instances[keyname] = new Log(keyname);
		return instances[keyname];
	
	}

	static void setPrecision(std::string keyname,int p){
		Log* log = getInstance(keyname);
		log->file.precision(p);
		
	}
	template <class T> static void write(std::string keyname, T value){


#ifdef LOGGING
		Log* log = getInstance(keyname);
		log->file << value<<" ";
#endif
	}

template <class T> static void writeNoSpace(std::string keyname, T value){


#ifdef LOGGING
		Log* log = getInstance(keyname);
		log->file << value;
#endif
	}

	template <class T> static void writeWithEnd(std::string keyname, T value){

#ifdef LOGGING
		Log* log = getInstance(keyname);
		log->file << value<<endl;
#endif
	}

	static void writeMatrix(std::string keyname, CvMat* temp){

#ifdef LOGGING
			Log* log = getInstance(keyname);

			for(int i=0;i < temp->rows;i++){
				
				for(int j=0;j < temp->cols;j++){
							
							log->file << cvmGet(temp,i,j)<<",";
							}
						//printf(";\n");
						//cout << endl;
				}
			//printf("\n");
#endif
	}

	static void writeMatrixND(std::string keyname, CvMatND* temp){

#ifdef LOGGING
			Log* log = getInstance(keyname);
			int nbins = temp->dims;
			for(int i=0; i < nbins; i++ ) {
				for(int j=0; j < nbins; j++ ) {
					for(int k=0; k < nbins; k++ ) {
						log->file << cvGetReal3D(temp,i,j,k)<<",";
					}
				}
			}
#endif
	}

	static void writeEnd(std::string keyname){

#ifdef LOGGING
		Log* log = getInstance(keyname);
		log->file <<endl;
#endif
	}

	
	
private:
	
	Log(){}
	Log(std::string filename){

#ifdef LOGGING

		file.setf(ios::fixed,ios::floatfield);
		file.precision(15);
		file.open (filename.c_str(), ios::out | ios::trunc  );

#endif
	}
	
	static std::map<std::string, Log*> instances;

	
	~Log(){
		file.close();
	}
};


#endif //LOGGING_H_
