#include <iostream>
#include <fstream>
#include <string>
#include <algorithm>
#include <cmath>
#include "opencv2/highgui/highgui.hpp"
#include "opencv2/stitching/stitcher.hpp"
#include "opencv2/stitching/detail/camera.hpp"
#include "opencv2/stitching/detail/motion_estimators.hpp"
#include "Logging.h"

using namespace std;
using namespace cv;

bool try_use_gpu = false;
vector<Mat> imgs;
vector<string> imgs_names;
string result_name = "result.jpg";

void printUsage();
int parseCmdArgs(int argc, char** argv);
cv::Mat composeRotationMat(const Vec3f& v);
Vec3f decomposeRotationMat(const cv::Mat& R);

int main(int argc, char* argv[])
{
    int retval = parseCmdArgs(argc, argv);
    if (retval) return -1;
    string mypano = "mypano.txt";
    Stitcher stitcher = Stitcher::createDefault(try_use_gpu);
    //stitcher.setPanoConfidenceThresh(1);
    //stitcher.setRegistrationResol(0.8);
    //stitcher.setWaveCorrection(false);
    cv::detail::BundleAdjusterBase* bundler = new detail::BundleAdjusterRay();//Reproj();
    //bundler->setTermCriteria(cvTermCriteria(CV_TERMCRIT_EPS + CV_TERMCRIT_ITER, 200, 0.01));
    stitcher.setBundleAdjuster(bundler);

    Stitcher::Status status = stitcher.estimateTransform(imgs);
    std::vector<detail::CameraParams> cameras = stitcher.cameras();
    std::vector<int> componentidx = stitcher.component();
    vector<int> allIdx; allIdx.resize(imgs_names.size());
    vector<int> output; output.resize(imgs_names.size()+componentidx.size());

    for (int i = 0; i < imgs_names.size(); ++i)
        allIdx[i] = i;
    
    std::sort (componentidx.begin(),componentidx.end());

    for (int i = 0; i < componentidx.size(); ++i)
    {
        cout <<"img idx " << componentidx[i]<<endl;
        detail::CameraParams cam = cameras[i];
        Log::writeNoSpace(mypano,"C:\\example\\");
        Log::writeWithEnd(mypano,imgs_names[componentidx[i]]);
        Log::write(mypano,imgs[componentidx[i]].cols);
        Log::writeWithEnd(mypano,imgs[componentidx[i]].rows);
        Log::writeEnd(mypano);

        Log::setPrecision(mypano,1);
        Log::write(mypano,1);
        Log::write(mypano,0);
        //Log::writeWithEnd(mypano,cam.ppx);
        Log::writeWithEnd(mypano,float(imgs[componentidx[i]].rows/2.0f));
        Log::write(mypano,0);
        Log::write(mypano,1);
        //Log::writeWithEnd(mypano,cam.ppy);
        Log::writeWithEnd(mypano,float(imgs[componentidx[i]].cols/2.0f));
        Log::write(mypano,0);
        Log::write(mypano,0);
        Log::writeWithEnd(mypano,1);
        Log::writeEnd(mypano);

        Vec3f angles = decomposeRotationMat(cam.R);
        cout << angles << endl;
        float swap = angles[0];
        angles[0] = angles[1];
        angles[1] = swap;
        cv::Mat R = composeRotationMat(angles);

        Log::setPrecision(mypano,6);
        Log::write(mypano,R.at<float>(0,0));
        Log::write(mypano,R.at<float>(0,1));
        //Log::writeWithEnd(mypano,ppx);
        Log::writeWithEnd(mypano,R.at<float>(0,2));
        Log::write(mypano,R.at<float>(1,0));
        Log::write(mypano,R.at<float>(1,1));
        //Log::writeWithEnd(mypano,ppy);
        Log::writeWithEnd(mypano,R.at<float>(1,2));
        Log::write(mypano,R.at<float>(2,0));
        Log::write(mypano,R.at<float>(2,1));
        Log::writeWithEnd(mypano,R.at<float>(2,2));
        Log::writeEnd(mypano);
        Log::setPrecision(mypano,3);
        Log::writeWithEnd(mypano,cam.focal);
        Log::writeEnd(mypano);

/*
Z:\home\micc\Documenti\autostitch\images\test\100-0024_img.jpg
568 758

1 0 378.5 
0 1 283.5 
0 0 1 

0.989768 -0.0039487 0.142633 
0.00390831 0.999992 0.000563158 
-0.142635 1.74769e-008 0.989775 

897.93

Z:\home\micc\Documenti\autostitch\images\test\100-0025_img.jpg
*/

        cout << "------------" << endl;
        cout << "Camera #"<< i << " " << imgs_names[componentidx[i]] << endl;
        cout << "focal "<< cam.focal << endl;
        cout << "ppx "<< cam.ppx << endl;
        cout << "ppy "<< cam.ppy << endl;
        cout << "Rotation "<< cam.R << endl;
        cout << "Aspect "<<  cam.aspect << endl;
        cout << "Translation "<< cam.t << endl;
    }
    std::vector<int>::iterator it;
    it = std::set_difference (allIdx.begin(), allIdx.end(), componentidx.begin(), componentidx.end(), output.begin());
    output.resize(it-output.begin());

    //! Writing missing images
    for (it=output.begin(); it!=output.end(); ++it){
        int i = *it;
        Log::writeNoSpace(mypano,"C:\\example\\");
        Log::writeWithEnd(mypano,imgs_names[i]);
        Log::write(mypano,imgs[i].cols);
        Log::writeWithEnd(mypano,imgs[i].rows);

        Log::setPrecision(mypano,1);
        Log::write(mypano,1);
        Log::write(mypano,0);
        //Log::writeWithEnd(mypano,cam.ppx);
        Log::writeWithEnd(mypano,float(imgs[i].rows/2.0f));
        Log::write(mypano,0);
        Log::write(mypano,1);
        //Log::writeWithEnd(mypano,cam.ppy);
        Log::writeWithEnd(mypano,float(imgs[i].cols/2.0f));
        Log::write(mypano,0);
        Log::write(mypano,0);
        Log::writeWithEnd(mypano,1);
        Log::writeEnd(mypano);

        Log::setPrecision(mypano,1);
        Log::write(mypano,0);
        Log::write(mypano,0);
        Log::writeWithEnd(mypano,0);
        Log::write(mypano,0);
        Log::write(mypano,0);
        Log::writeWithEnd(mypano,0);
        Log::write(mypano,0);
        Log::write(mypano,0);
        Log::writeWithEnd(mypano,0);
        Log::writeEnd(mypano);
        Log::setPrecision(mypano,1);
        Log::writeWithEnd(mypano,0);
        Log::writeEnd(mypano);
    }

    if (status != Stitcher::OK)
    {
        cout << "Can't stitch images, error code = " << int(status) << endl;
        return -1;
    }
    else cout << "Transformations are estimated." << endl;

    //imwrite(result_name, pano);
    return 0;
}


cv::Mat composeRotationMat(const Vec3f& v){
    cv::Mat_<float> R,X,Y,Z;
    float x = v[0];
    float y = v[1];
    float z = v[2];
    X = Mat::eye(3, 3, CV_32F);
    Y = Mat::eye(3, 3, CV_32F);
    Z = Mat::eye(3, 3, CV_32F);

    X.at<float>(1,1) = cos(x);
    X.at<float>(1,2) = -sin(x);
    X.at<float>(2,1) = sin(x);
    X.at<float>(2,2) = cos(x);

    Y.at<float>(0,0) = cos(y);
    Y.at<float>(0,2) = sin(y);
    Y.at<float>(2,0) = -sin(y);
    Y.at<float>(2,2) = cos(y);

    Z.at<float>(0,0) = cos(z);
    Z.at<float>(0,1) = -sin(z);
    Z.at<float>(1,0) = sin(z);
    Z.at<float>(1,1) = cos(z);

    cout<< "***X***" <<endl;
    cout << X <<endl;

    R = Z*Y*X;


    return R;
}

Vec3f decomposeRotationMat(const cv::Mat& R){
    /*function [x,y,z] = decompose_rotation(R)
    %x = atan2(R(3,2), R(3,3));
    %y = atan2(-R(3,1), sqrt(R(3,2)*R(3,2) + R(3,3)*R(3,3)));
    %z = atan2(R(2,1), R(1,1));
    */
    float x = atan2( R.at<float>(2,1), R.at<float>(2,2) );
    float y = atan2( -R.at<float>(2,0),
                     sqrt( pow(R.at<float>(2,1),2) +
                     pow(R.at<float>(2,2),2) ) );
    float z = atan2( R.at<float>(1,0), R.at<float>(0,0) );
    return Vec3f(x,y,z);
}


void printUsage()
{
    cout <<
        "Usage: stitching_micc img1 img2 [...imgN]\n\n"
        "The default output is 'mypano.txt'.\n";
}


int parseCmdArgs(int argc, char** argv)
{
    if (argc == 1)
    {
        printUsage();
        return -1;
    }
    for (int i = 1; i < argc; ++i)
    {
        if (string(argv[i]) == "--help" || string(argv[i]) == "/?")
        {
            printUsage();
            return -1;
        }
        else if (string(argv[i]) == "--try_use_gpu")
        {
            if (string(argv[i + 1]) == "no")
                try_use_gpu = false;
            else if (string(argv[i + 1]) == "yes")
                try_use_gpu = true;
            else
            {
                cout << "Bad --try_use_gpu flag value\n";
                return -1;
            }
            i++;
        }
        else if (string(argv[i]) == "--output")
        {
            result_name = argv[i + 1];
            i++;
        }
        else
        {
            Mat img = imread(argv[i]);
            if (img.empty())
            {
                cout << "Can't read image '" << argv[i] << "'\n";
                return -1;
            }
            imgs.push_back(img);
            imgs_names.push_back(string(argv[i]));
        }
    }
    return 0;
}
